#include "vmdevice.h"
#include <stdint.h>
#include <stdio.h>
#include "net.h"
#include "ds1621.h"
#include "mb.h"
#include "mb_ds1621_if.h"

#define DS1621_ADDR               0x92
#define DS1621_MAX_COILS          8
#define DS1621_START_HOLD_REG     0
#define DS1621_END_HOLD_REG       2
#define DS1621_START_INPUT_REG    4
#define DS1621_END_INPUT_REG      8


#define VENDORNAME_LEN (sizeof("Coptonix GmbH")-1)
#define VendorName          0x00,\
                            VENDORNAME_LEN,\
                            'C','o','p','t','o','n','i','x',' ','G','m','b','H'

#define PRODUCT_LEN (sizeof("MB DS1621")-1)
#define ProductCode         0x01,\
                            PRODUCT_LEN,\
                            'M','B',' ','D','S','1','6','2','1'

#define REVISION_LEN (sizeof("V1.00")-1)
#define MajorMinorRevision  0x02,\
                            REVISION_LEN,\
                            'V','1','.','0','0'

const uint8_t BasicDeviceObjects [] =
{
    VendorName,
    ProductCode,
    MajorMinorRevision
};

uint8_t mb_read_coils(void)
{
    uint8_t value;
    uint8_t ret;
    uint8_t mask = 0;
    uint8_t i;
    uint16_t startaddr;
    uint16_t numcoils;

    startaddr = u16swap(modbus.request.u.RdCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.RdCoils.NumCoils);
#if (USE_PRINT_U16 == 1)
    print_u16("StartAddr 0x%x\r",startaddr);
    print_u16("NumCoils 0x%x\r",numcoils);
#endif
    if ((numcoils < MB_RD_MIN_NUM_COILS) || (numcoils >MB_RD_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (DS1621_MAX_COILS-1)) || (numcoils > (DS1621_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (ds1621_get_config(DS1621_ADDR, &value)) {
        value >>= startaddr;
        for(i=0;i<numcoils;i++){
            mask += (1<<i);
        }
        value &= mask;
        modbus.response.u.RdCoils.Count = 1;            // "Byte Count" -> 8Bits config value, therefore only a max/min of one byte
        modbus.response.u.RdCoils.Status[0] = value;    // "Coil Status"
        mb_set_data_length(2);                          // 2 bytes: 1x "Byte Count" and 1x "Config"
        ret = MB_OK;
    } else ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    return(ret);
}

uint8_t mb_write_single_coil(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t coilvalue;
    uint8_t value;

    addr = u16swap(modbus.request.u.WrSingleCoil.Addr);
    coilvalue = u16swap(modbus.request.u.WrSingleCoil.Value);

#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",addr);
    print_u16("Value 0x%x\r",coilvalue);
    //printf("addr = 0x%x ; Value = 0x%x\r",addr,coilvalue);
#endif

    if (!((coilvalue == MB_COIL_LOW) || (coilvalue == MB_COIL_HIGH))){
        return(MB_ERR_ILLEGAL_DATA_VALUE);
    }

    if (addr > (DS1621_MAX_COILS-1))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    if (ds1621_get_config(DS1621_ADDR, &value)) {

        if (coilvalue == MB_COIL_HIGH) {
            value |= (1 << ((uint8_t)(addr & 0x07)));
        } else {
            value &= ~(1 << ((uint8_t)(addr & 0x07)));
        }

        if (ds1621_set_config(DS1621_ADDR, value)) {
            modbus.response.u.WrSingleCoil.Addr = modbus.request.u.WrSingleCoil.Addr;
            modbus.response.u.WrSingleCoil.Value = modbus.request.u.WrSingleCoil.Value;
            mb_set_data_length(4);
            ret = MB_OK;
        }
    }
    return(ret);
}

uint8_t mb_write_single_register(void)
{
    uint8_t ret;
    uint16_t addr;
    uint16_t value;

    addr = u16swap(modbus.request.u.WrSingleReg.Addr);
    value = u16swap(modbus.request.u.WrSingleReg.Value);
#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",addr);
    print_u16("Value 0x%x\r",value);
    //printf("addr = 0x%x ; Value = 0x%x\r",addr,value);
#endif

    modbus.response.u.WrSingleReg.Addr = modbus.request.u.WrSingleReg.Addr;
    modbus.response.u.WrSingleReg.Value = modbus.request.u.WrSingleReg.Value;
    mb_set_data_length(4);
    ret = MB_OK;

    switch(addr) {
    case 0x00:
        if (!ds1621_set_config(DS1621_ADDR, (uint8_t)(value & 0x00FF))) {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
        break;
   case 0x01:
        if (!ds1621_set_TH(DS1621_ADDR, value)) {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
        break;
    case 0x02:
        if (!ds1621_set_TL(DS1621_ADDR, value)) {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
        break;
    case 0x03: // write-only register
        if (value == 0xFFFF) {
            if (!ds1621_start_convert(DS1621_ADDR)) {
                ret = MB_ERR_SLAVE_DEVICE_FAILURE;
            }
        } else {
            if (!ds1621_stop_convert(DS1621_ADDR)) {
                ret = MB_ERR_SLAVE_DEVICE_FAILURE;
            }
        }
        break;
    default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
        break;
    }
    return(ret);
}

uint8_t mb_write_multiple_coils(void)
{
    uint8_t ret;
    uint16_t startaddr;
    uint16_t numcoils;
    uint8_t coils_state;
    uint8_t value;
    uint8_t mask = 0;
    uint8_t i;

    startaddr = u16swap(modbus.request.u.WrMultCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.WrMultCoils.NumCoils);

#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",startaddr);
    print_u16("NumCoils 0x%x\r",numcoils);
    //printf("Addr = 0x%x ; NumCoils = 0x%x\r",addr,numcoils);
#endif
    if ((numcoils < MB_WR_MIN_NUM_COILS) || (numcoils > MB_WR_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.src_socket.rxlen < MB_WRITE_MIN_PACKET_SIZE)
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.request.u.WrMultCoils.Count != (modbus.src_socket.rxlen - (MB_WRITE_MIN_PACKET_SIZE-1)))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (DS1621_MAX_COILS-1)) || (numcoils > (DS1621_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    coils_state = modbus.request.u.WrMultCoils.Value[0];

    if (ds1621_get_config(DS1621_ADDR, &value)) {
        for(i=0;i<numcoils;i++){
            mask += (1<<i);
        }
        mask <<= startaddr;
        value &= ~mask;
        coils_state <<= startaddr;
        coils_state &= mask;
        value |= coils_state;

        if (ds1621_set_config(DS1621_ADDR, value)) {
            modbus.response.u.WrMultCoils.StartAddr = modbus.request.u.WrMultCoils.StartAddr;
            modbus.response.u.WrMultCoils.NumCoils = modbus.request.u.WrMultCoils.NumCoils;
            mb_set_data_length(4);
            ret = MB_OK;
        } else {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_mask_write_register(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint8_t value;
    uint16_t addr;
    uint16_t andmask;
    uint16_t ormask;

    addr = u16swap(modbus.request.u.MaskWrReg.Addr);
    andmask = u16swap(modbus.request.u.MaskWrReg.AndMask);
    ormask = u16swap(modbus.request.u.MaskWrReg.OrMask);
#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",addr);
    print_u16("AND 0x%x\r",andmask);
    print_u16("OR 0x%x\r",ormask);
#endif
    if (addr != 0) //only to be used with register 0 // CONFIG Regsiter
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (ds1621_get_config(DS1621_ADDR, &value)) {
        value = (value & andmask) | (ormask & (~andmask));
        if (ds1621_set_config(DS1621_ADDR, value)) {
            modbus.response.u.MaskWrReg.Addr = modbus.request.u.MaskWrReg.Addr;
            modbus.response.u.MaskWrReg.AndMask = modbus.request.u.MaskWrReg.AndMask;
            modbus.response.u.MaskWrReg.OrMask = modbus.request.u.MaskWrReg.OrMask;
            mb_set_data_length(6);
            ret = MB_OK;
        } else {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_read_holding_registers(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t numregs;
    uint16_t len;
    uint8_t i;
    uint8_t u8value;
    uint16_t u16value;

    addr = u16swap(modbus.request.u.RdHoldingRegs.StartAddr);
    numregs = u16swap(modbus.request.u.RdHoldingRegs.NumRegisters);
#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",addr);
    print_u16("NumRegs 0x%x\r",numregs);
#endif
    if ((numregs < MB_RD_MIN_NUM_REGS) || (numregs > MB_RD_MAX_NUM_REGS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((addr > DS1621_END_HOLD_REG) || (numregs > ((DS1621_END_HOLD_REG+1) - addr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    modbus.response.u.RdHoldingRegs.Count = (uint8_t)(numregs * 2);
    len = 1;
    ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    for (i = addr; i<(addr + numregs); i++) {
        switch(i) {
        case 0x00:
            if (ds1621_get_config(DS1621_ADDR, &u8value)) {
                modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap((uint16_t)u8value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x01:
            if (ds1621_get_TH(DS1621_ADDR, &u16value)) {
                modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap(u16value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x02:
            if (ds1621_get_TL(DS1621_ADDR, &u16value)) {
                modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap(u16value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
                 break;
        }
        if(ret != MB_OK)
            break;
    }
    return(ret);
}

uint8_t mb_read_input_registers(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t numregs;
    uint16_t len;
    uint8_t i;
    uint8_t u8value;
    uint16_t u16value;


    static FL_INT_T  temp;
    static uint8_t last_result = 0;

    addr = u16swap(modbus.request.u.RdInputRegs.StartAddr);
    numregs = u16swap(modbus.request.u.RdInputRegs.NumRegisters);
#if (USE_PRINT_U16 == 1)
    print_u16("Addr 0x%x\r",addr);
    print_u16("NumRegs 0x%x\r",numregs);
#endif
    if ((numregs < MB_RD_MIN_NUM_REGS) || (numregs > MB_RD_MAX_NUM_REGS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((addr < DS1621_START_INPUT_REG) || (addr > DS1621_END_INPUT_REG)) {
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);
    }

    if (numregs > ((DS1621_END_INPUT_REG - DS1621_START_INPUT_REG + 1) - (addr - DS1621_START_INPUT_REG)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    modbus.response.u.RdInputRegs.Count = (uint8_t)(numregs * 2);
    len = 1;
    ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    for (i = addr; i<(addr + numregs); i++) {
        switch(i) {
        case 0x04:
            if (ds1621_read_temp(DS1621_ADDR, &u16value)) {
                modbus.response.u.RdInputRegs.Value[i - addr] = u16swap(u16value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x05:
            if (ds1621_read_counter(DS1621_ADDR, &u8value)) {
                modbus.response.u.RdInputRegs.Value[i - addr] = u16swap((uint16_t)u8value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x06:
            if (ds1621_read_slope(DS1621_ADDR, &u8value)) {
                modbus.response.u.RdInputRegs.Value[i - addr] = u16swap((uint16_t)u8value);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x07:
            last_result = ds1621_read_hrtemp(DS1621_ADDR, &temp.fl);
            if (last_result) {
                modbus.response.u.RdInputRegs.Value[i - addr] = u16swap(temp.u16[1]);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        case 0x08:
            if (last_result) {
                modbus.response.u.RdInputRegs.Value[i - addr] = u16swap(temp.u16[0]);
                len += 2;
                mb_set_data_length(len);
                ret = MB_OK;
            }
            break;
        default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
                 break;
        }
        if (ret != MB_OK)
            break;
    }
    return(ret);
}

void mb_device_init(void)
{
    mb_init(BasicDeviceObjects, sizeof(BasicDeviceObjects), CH_VMTCP);

    mb_register_function(MB_CMD_READ_COILS, mb_read_coils);
    //mb_register_mb_function(MB_CMD_READ_DISCRETE_INPUTS, NULL);
    mb_register_function(MB_CMD_READ_HOLDING_REGISTERS, mb_read_holding_registers);
    mb_register_function(MB_CMD_READ_INPUT_REGISTERS, mb_read_input_registers);
    mb_register_function(MB_CMD_WRITE_SINGLE_COIL, mb_write_single_coil);
    mb_register_function(MB_CMD_WRITE_SINGLE_REGISTER, mb_write_single_register);
    mb_register_function(MB_CMD_WRITE_MULTIPLE_COILS, mb_write_multiple_coils);
    //mb_register_function(MB_CMD_WRITE_MULTIPLE_REGISTERS, NULL);
    mb_register_function(MB_CMD_WRITE_MASK_REGISTER, mb_mask_write_register);
    //mb_register_function(MB_CMD_WRITE_READ_MULTIPLE_REGISTERS, NULL);
}
