#include "vmdevice.h"
#include <stdint.h>
#include <stdio.h>
#include "net.h"
#include "ltc1669.h"
#include "mb.h"
#include "mb_ltc1669_if.h"

#define LTC1669_ADDR        0x40
#define LTC1669_MAX_COILS   3
#define LTC1669_MAX_REGS    1

#define VENDORNAME_LEN (sizeof("Coptonix GmbH")-1)
#define VendorName          0x00,\
                            VENDORNAME_LEN,\
                            'C','o','p','t','o','n','i','x',' ','G','m','b','H'

#define PRODUCT_LEN (sizeof("MB LTC1669")-1)
#define ProductCode         0x01,\
                            PRODUCT_LEN,\
                            'M','B',' ','L','T','C','1','6','6','9'

#define REVISION_LEN (sizeof("V1.00")-1)
#define MajorMinorRevision  0x02,\
                            REVISION_LEN,\
                            'V','1','.','0','0'

const uint8_t BasicDeviceObjects [] =
{
    VendorName,
    ProductCode,
    MajorMinorRevision
};

uint8_t mb_read_coils(void)
{
    uint8_t value;
    uint8_t ret;
    uint8_t mask = 0;
    uint8_t i;
    uint16_t startaddr;
    uint16_t numcoils;

    startaddr = u16swap(modbus.request.u.RdCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.RdCoils.NumCoils);

    if ((numcoils < MB_RD_MIN_NUM_COILS) || (numcoils > MB_RD_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (LTC1669_MAX_COILS-1)) || (numcoils > (LTC1669_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    value = read_dac_cmd();
    value >>= startaddr;
    for(i=0;i<numcoils;i++){
        mask += (1<<i);
    }
    value &= mask;
    modbus.response.u.RdCoils.Count = 1;            // "Byte Count" -> 3Bits , therefore only a max/min of one byte
    modbus.response.u.RdCoils.Status[0] = value;    // "Coil Status"
    mb_set_data_length(2);                          // 2 bytes: 1x "Byte Count" and 1x Byte
    ret = MB_OK;

    return(ret);
}

uint8_t mb_write_single_coil(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t coilvalue;
    uint8_t value;

    addr = u16swap(modbus.request.u.WrSingleCoil.Addr);
    coilvalue = u16swap(modbus.request.u.WrSingleCoil.Value);

    if (!((coilvalue == MB_COIL_LOW) || (coilvalue == MB_COIL_HIGH))){
        return(MB_ERR_ILLEGAL_DATA_VALUE);
    }

    if (addr > (LTC1669_MAX_COILS-1))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (coilvalue == MB_COIL_HIGH) {
        value = 1;
    } else {
        value = 0;
    }

    if (set_dac_cmd_bit(LTC1669_ADDR, addr, value)) {
        modbus.response.u.WrSingleCoil.Addr = modbus.request.u.WrSingleCoil.Addr;
        modbus.response.u.WrSingleCoil.Value = modbus.request.u.WrSingleCoil.Value;
        mb_set_data_length(4);
        ret = MB_OK;
    } else ret = MB_ERR_SLAVE_DEVICE_FAILURE;

    return(ret);
}

uint8_t mb_write_single_register(void)
{
    uint8_t ret;
    uint16_t addr;
    uint16_t value;

    addr = u16swap(modbus.request.u.WrSingleReg.Addr);
    value = u16swap(modbus.request.u.WrSingleReg.Value);

    modbus.response.u.WrSingleReg.Addr = modbus.request.u.WrSingleReg.Addr;
    modbus.response.u.WrSingleReg.Value = modbus.request.u.WrSingleReg.Value;
    mb_set_data_length(4);
    ret = MB_OK;

    switch(addr) {
    case 0x00:
        if (!write_dac_value(LTC1669_ADDR, value)) {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
        break;
    default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
        break;
    }
    return(ret);
}

uint8_t mb_write_multiple_coils(void)
{
    uint8_t ret;
    uint16_t startaddr;
    uint16_t numcoils;
    uint8_t coils_state;
    uint8_t value;
    uint8_t mask = 0;
    uint8_t i;

    startaddr = u16swap(modbus.request.u.WrMultCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.WrMultCoils.NumCoils);

    if ((numcoils < MB_WR_MIN_NUM_COILS) || (numcoils > MB_WR_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.src_socket.rxlen < MB_WRITE_MIN_PACKET_SIZE)
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.request.u.WrMultCoils.Count != (modbus.src_socket.rxlen - (MB_WRITE_MIN_PACKET_SIZE-1)))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (LTC1669_MAX_COILS-1)) || (numcoils > (LTC1669_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    coils_state = modbus.request.u.WrMultCoils.Value[0];

    value = read_dac_cmd();

    for(i=0;i<numcoils;i++){
        mask += (1<<i);
    }
    mask <<= startaddr;
    value &= ~mask;
    coils_state <<= startaddr;
    coils_state &= mask;
    value |= coils_state;

    if (write_dac_cmd(LTC1669_ADDR, value)) {
        modbus.response.u.WrMultCoils.StartAddr = modbus.request.u.WrMultCoils.StartAddr;
        modbus.response.u.WrMultCoils.NumCoils = modbus.request.u.WrMultCoils.NumCoils;
        mb_set_data_length(4);
        ret = MB_OK;
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_read_holding_registers(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t numregs;
    uint16_t len;
    uint8_t i;
    uint16_t u16value;

    addr = u16swap(modbus.request.u.RdHoldingRegs.StartAddr);
    numregs = u16swap(modbus.request.u.RdHoldingRegs.NumRegisters);

    if ((numregs < MB_RD_MIN_NUM_REGS) || (numregs > MB_RD_MAX_NUM_REGS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((addr > (LTC1669_MAX_REGS-1)) || (numregs > (LTC1669_MAX_REGS - addr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    modbus.response.u.RdHoldingRegs.Count = (uint8_t)(numregs * 2);
    len = 1;
    //ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    for (i = addr; i<(addr + numregs); i++) {
        switch(i) {
        case 0x00:
            u16value = read_dac_value();
            modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap(u16value);
            len += 2;
            mb_set_data_length(len);
            ret = MB_OK;
            break;
        default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
                 return(ret);
                 break;
        }
    }
    return(ret);
}

void mb_device_init(void)
{
    mb_init(BasicDeviceObjects, sizeof(BasicDeviceObjects), CH_VMTCP);

    mb_register_function(MB_CMD_READ_COILS, mb_read_coils);
    //mb_register_mb_function(MB_CMD_READ_DISCRETE_INPUTS, NULL);
    mb_register_function(MB_CMD_READ_HOLDING_REGISTERS, mb_read_holding_registers);
    //mb_register_function(MB_CMD_READ_INPUT_REGISTERS, mb_read_input_registers);
    mb_register_function(MB_CMD_WRITE_SINGLE_COIL, mb_write_single_coil);
    mb_register_function(MB_CMD_WRITE_SINGLE_REGISTER, mb_write_single_register);
    mb_register_function(MB_CMD_WRITE_MULTIPLE_COILS, mb_write_multiple_coils);
    //mb_register_function(MB_CMD_WRITE_MULTIPLE_REGISTERS, NULL);
    //mb_register_function(MB_CMD_WRITE_MASK_REGISTER, mb_mask_write_register);
    //mb_register_function(MB_CMD_WRITE_READ_MULTIPLE_REGISTERS, NULL);
}
