#include "vmdevice.h"
#include <stdint.h>
#include <stdio.h>
#include "net.h"
#include "pcf8574x.h"
#include "mb.h"
#include "mb_pcf8574_if.h"

#define PCF8574_MAX_COILS   8
#define PCF8574_MAX_REGS    2

#define VENDORNAME_LEN (sizeof("Coptonix GmbH")-1)
#define VendorName          0x00,\
                            VENDORNAME_LEN,\
                            'C','o','p','t','o','n','i','x',' ','G','m','b','H'

#define PRODUCT_LEN (sizeof("MB IO-EXPANDER")-1)
#define ProductCode         0x01,\
                            PRODUCT_LEN,\
                            'M','B',' ','I','O','-','E','P','A','N','D','E','R'

#define REVISION_LEN (sizeof("V1.00")-1)
#define MajorMinorRevision  0x02,\
                            REVISION_LEN,\
                            'V','1','.','0','0'

static uint8_t io_slaveaddr;

const uint8_t BasicDeviceObjects [] =
{
    VendorName,
    ProductCode,
    MajorMinorRevision
};

uint8_t mb_read_coils(void)
{
    uint8_t value;
    uint8_t ret;
    uint8_t mask = 0;
    uint8_t i;
    uint16_t startaddr;
    uint16_t numcoils;

    //Convert Endianness (big-endian to little-endian)
    startaddr = u16swap(modbus.request.u.RdCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.RdCoils.NumCoils);

#ifdef DEBUG_PRINT
    print_u16("StartAddr 0x%x\r",startaddr);
    print_u16("NumCoils 0x%x\r",numcoils);
#endif

    if ((numcoils < MB_RD_MIN_NUM_COILS) || (numcoils > MB_RD_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (PCF8574_MAX_COILS-1)) || (numcoils > (PCF8574_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (io_readport(io_slaveaddr, &value)) {
        value >>= startaddr;
        for(i=0;i<numcoils;i++){
            mask += (1<<i);
        }
        value &= mask;
        modbus.response.u.RdCoils.Count = 1;            // 1 byte -> "Byte Count" -> 8Bits IO-Expander, therefore only one byte
        modbus.response.u.RdCoils.Status[0] = value;    // 1 byte -> "Coil Status"
        mb_set_data_length(2);                          // 2 bytes: 1x "Byte Count" and 1x "Coil Status"
        ret = MB_OK;
    } else ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    return(ret);
}

uint8_t mb_write_single_coil(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t coilvalue;
    uint8_t coilstate;

    //Convert Endianness (big-endian to little-endian)
    addr = u16swap(modbus.request.u.WrSingleCoil.Addr);
    coilvalue = u16swap(modbus.request.u.WrSingleCoil.Value);

#ifdef DEBUG_PRINT
    print_u16("Addr 0x%x\r",addr);
    print_u16("Value 0x%x\r",coilvalue);
    //printf("addr = 0x%x ; Value = 0x%x\r",addr,coilvalue);
#endif

    if (coilvalue == MB_COIL_LOW) {
        coilstate = 0x00;
    } else if (coilvalue == MB_COIL_HIGH) {
        coilstate = 0x01;
    } else {
        return(MB_ERR_ILLEGAL_DATA_VALUE);
    }

   if (addr > (PCF8574_MAX_COILS-1))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (io_setpinstate(io_slaveaddr, addr,coilstate)) {
        //Copy data from REQUEST to RESPONSE
        modbus.response.u.WrSingleCoil.Addr = modbus.request.u.WrSingleCoil.Addr;   //2 bytes
        modbus.response.u.WrSingleCoil.Value = modbus.request.u.WrSingleCoil.Value; //2 bytes
        mb_set_data_length(4);                                                      //4 bytes total
        ret = MB_OK;
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_write_single_register(void)
{
    uint8_t ret;
    uint16_t addr;
    uint16_t value;

    //Convert Endianness (big-endian to little-endian)
    addr = u16swap(modbus.request.u.WrSingleReg.Addr);
    value = u16swap(modbus.request.u.WrSingleReg.Value);
#ifdef DEBUG_PRINT
    print_u16("Addr 0x%x\r",addr);
    print_u16("Value 0x%x\r",value);
    //printf("addr = 0x%x ; Value = 0x%x\r",addr,value);
#endif
    modbus.response.u.WrSingleReg.Addr = modbus.request.u.WrSingleReg.Addr;
    modbus.response.u.WrSingleReg.Value = modbus.request.u.WrSingleReg.Value;
    mb_set_data_length(4);
    ret = MB_OK;
    switch(addr) {
    case 0 :    io_slaveaddr = (uint8_t)(value & 0x00FE);
                break;
    case 1 :    if (!io_writeport(io_slaveaddr, (uint8_t)(value & 0x00FF))) {
                    ret = MB_ERR_SLAVE_DEVICE_FAILURE;
                }
                break;
    default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS; break;
    }
    return(ret);
}

uint8_t mb_write_multiple_coils(void)
{
    uint8_t ret;
    uint16_t startaddr;
    uint16_t numcoils;
    uint8_t coils_state;
    uint8_t value;
    uint8_t mask = 0;
    uint8_t i;

    //Convert Endianness (big-endian to little-endian)
    startaddr = u16swap(modbus.request.u.WrMultCoils.StartAddr);
    numcoils = u16swap(modbus.request.u.WrMultCoils.NumCoils);

#ifdef DEBUG_PRINT
    print_u16("Addr 0x%x\r",startaddr);
    print_u16("NumCoils 0x%x\r",numcoils);
    //printf("Addr = 0x%x ; NumCoils = 0x%x\r",addr,numcoils);
#endif
    if ((numcoils < MB_WR_MIN_NUM_COILS) || (numcoils > MB_WR_MAX_NUM_COILS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.src_socket.rxlen < MB_WRITE_MIN_PACKET_SIZE)
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if (modbus.request.u.WrMultCoils.Count != (modbus.src_socket.rxlen - (MB_WRITE_MIN_PACKET_SIZE-1)))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((startaddr > (PCF8574_MAX_COILS-1)) || (numcoils > (PCF8574_MAX_COILS - startaddr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    coils_state = modbus.request.u.WrMultCoils.Value[0];

    if (io_readport(io_slaveaddr, &value)) {
        for(i=0;i<numcoils;i++){
            mask += (1<<i);
        }
        mask <<= startaddr;
        value &= ~mask;
        coils_state <<= startaddr;
        coils_state &= mask;
        value |= coils_state;

        if (io_writeport(io_slaveaddr, value)) {
            modbus.response.u.WrMultCoils.StartAddr = modbus.request.u.WrMultCoils.StartAddr;
            modbus.response.u.WrMultCoils.NumCoils = modbus.request.u.WrMultCoils.NumCoils;
            mb_set_data_length(4);
            ret = MB_OK;
        } else {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_mask_write_register(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint8_t value;
    uint16_t addr;
    uint16_t andmask;
    uint16_t ormask;

    addr = u16swap(modbus.request.u.MaskWrReg.Addr);
    andmask = u16swap(modbus.request.u.MaskWrReg.AndMask);
    ormask = u16swap(modbus.request.u.MaskWrReg.OrMask);
#ifdef DEBUG_PRINT
    print_u16("Addr 0x%x\r",addr);
    print_u16("AND 0x%x\r",andmask);
    print_u16("OR 0x%x\r",ormask);
#endif
    if (addr != 1) //only to be used with register 1 (Expander IO)
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (io_readport(io_slaveaddr, &value)) {
        value = (value & andmask) | (ormask & (~andmask));
        if (io_writeport(io_slaveaddr, value)) {
            modbus.response.u.MaskWrReg.Addr = modbus.request.u.MaskWrReg.Addr;
            modbus.response.u.MaskWrReg.AndMask = modbus.request.u.MaskWrReg.AndMask;
            modbus.response.u.MaskWrReg.OrMask = modbus.request.u.MaskWrReg.OrMask;
            mb_set_data_length(6);
            ret = MB_OK;
        } else {
            ret = MB_ERR_SLAVE_DEVICE_FAILURE;
        }
    } else {
        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
    }
    return(ret);
}

uint8_t mb_read_holding_registers(void)
{
    uint8_t ret = MB_ERR_UNKNOWN;
    uint16_t addr;
    uint16_t numregs;
    uint16_t len;
    uint8_t i;
    uint8_t value;

    addr = u16swap(modbus.request.u.RdHoldingRegs.StartAddr);
    numregs = u16swap(modbus.request.u.RdHoldingRegs.NumRegisters);
#ifdef DEBUG_PRINT
    print_u16("Addr 0x%x\r",addr);
    print_u16("NumRegs 0x%x\r",numregs);
#endif
    if ((numregs < MB_RD_MIN_NUM_REGS) || (numregs > MB_RD_MAX_NUM_REGS))
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    if ((addr > (PCF8574_MAX_REGS-1)) || (numregs > (PCF8574_MAX_REGS - addr)))
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    modbus.response.u.RdHoldingRegs.Count = (uint8_t)(numregs * 2);
    len = 1;
    for (i = addr; i<(addr + numregs); i++) {
        switch(i) {
        case 0 :    modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap((uint16_t)io_slaveaddr);
                    len += 2;
                    mb_set_data_length(len);
                    ret = MB_OK;
                    break;
        case 1 :    if (io_readport(io_slaveaddr, &value)) {
                        modbus.response.u.RdHoldingRegs.Value[i - addr] = u16swap((uint16_t)value);
                        len += 2;
                        mb_set_data_length(len);
                        ret = MB_OK;
                    } else {
                        ret = MB_ERR_SLAVE_DEVICE_FAILURE;
                    }
                    break;
        default: ret = MB_ERR_ILLEGAL_DATA_ADDRESS;
                 break;
        }
        if (ret != MB_OK)
            break;
    }

    return(ret);
}

void mb_device_init(void)
{
    mb_init(BasicDeviceObjects, sizeof(BasicDeviceObjects), CH_VMTCP);

    io_slaveaddr = 0x40;

    mb_register_function(MB_CMD_READ_COILS, mb_read_coils);
    //mb_register_function(MB_CMD_READ_DISCRETE_INPUTS, NULL);
    mb_register_function(MB_CMD_READ_HOLDING_REGISTERS, mb_read_holding_registers);
    //mb_register_function(MB_CMD_READ_INPUT_REGISTERS, NULL);
    mb_register_function(MB_CMD_WRITE_SINGLE_COIL, mb_write_single_coil);
    mb_register_function(MB_CMD_WRITE_SINGLE_REGISTER, mb_write_single_register);
    mb_register_function(MB_CMD_WRITE_MULTIPLE_COILS, mb_write_multiple_coils);
    //mb_register_function(MB_CMD_WRITE_MULTIPLE_REGISTERS, NULL);
    mb_register_function(MB_CMD_WRITE_MASK_REGISTER, mb_mask_write_register);
    //mb_register_function(MB_CMD_WRITE_READ_MULTIPLE_REGISTERS, NULL);
}
