/*--------------------------------------------------------------------------------
   mb.h: Modbus Library for the Virtual Machine 8051 / LAN I2C Adapter VM

   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   -------------------------------------------------------------------------------
   Date:    2022-07-19
   Version: 1.00
   Comment: Initial release
----------------------------------------------------------------------------------*/

#ifndef MB_H_INCLUDED
#define MB_H_INCLUDED

#include "net.h"

#define MB_RD_MIN_NUM_REGS                  0x0001
#define MB_RD_MAX_NUM_REGS                  0x007D
#define MB_RD_MIN_NUM_COILS                 0x0001
#define MB_RD_MAX_NUM_COILS                 0x07D0
#define MB_WR_MIN_NUM_REGS                  0x0001
#define MB_WR_MAX_NUM_REGS                  0x007B
#define MB_WR_MIN_NUM_COILS                 0x0001
#define MB_WR_MAX_NUM_COILS                 0x07B0
#define MB_MIN_NUM_MULTI_REGS               0x0001
#define MB_WR_MAX_NUM_MULTI_REGS            0x0079
#define MB_RD_MAX_NUM_MULTI_REGS            0x007D
#define MB_COIL_LOW                         0x0000
#define MB_COIL_HIGH                        0xFF00
#define MB_WRITE_MIN_PACKET_SIZE            14

#define MB_OK                               0x00
#define MB_ERR_ILLEGAL_FUNCTION             0x01
#define MB_ERR_ILLEGAL_DATA_ADDRESS         0x02
#define MB_ERR_ILLEGAL_DATA_VALUE           0x03
#define MB_ERR_SLAVE_DEVICE_FAILURE         0x04
#define MB_ERR_ACKNOWLEDGE                  0x05
#define MB_ERR_SLAVE_DEVICE_BUSY            0x06
#define MB_ERR_NEGATIVE_ACKNOWLEDGE         0x07
#define MB_ERR_MEMORY_PARITY_ERROR          0x08
#define MB_ERR_GATYWAY_PATH_UNAVAILABLE     0x0A
#define MB_ERR_GATEWAY_TARGET_DEVICE_FAILED_TO_RESPOND   0x0B
#define MB_ERR_UNKNOWN                      0xFF

#define MB_CMD_READ_COILS                       0x01
#define MB_CMD_READ_DISCRETE_INPUTS             0x02
#define MB_CMD_READ_HOLDING_REGISTERS           0x03
#define MB_CMD_READ_INPUT_REGISTERS             0x04
#define MB_CMD_WRITE_SINGLE_COIL                0x05
#define MB_CMD_WRITE_SINGLE_REGISTER            0x06
#define MB_CMD_READ_EXCEPTION_STATUS            0x07 // Serial Line only
#define MB_CMD_READ_DIAGNOSTICS                 0x08 // Serial Line only
#define MB_CMD_GET_COMM_EVENT_COUNTER           0x0B // Serial Line only
#define MB_CMD_GET_COMM_EVENT_LOG               0x0C // Serial Line only
#define MB_CMD_WRITE_MULTIPLE_COILS             0x0F
#define MB_CMD_WRITE_MULTIPLE_REGISTERS         0x10
#define MB_CMD_REPORT_SERVER_ID                 0x11 // Serial Line only
#define MB_CMD_READ_FILE_RECORD                 0x14
#define MB_CMD_WRITE_FILE_RECORD                0x15
#define MB_CMD_WRITE_MASK_REGISTER              0x16
#define MB_CMD_WRITE_READ_MULTIPLE_REGISTERS    0x17
#define MB_CMD_READ_FIFO_QUEUE                  0x18
#define MB_CMD_READ_DEVICE_INFORMATION          0x2B

typedef struct __MB_FUNCTIONS_T
{
   uint8_t (*read_coils) (void);
   uint8_t (*read_discrete_inputs) (void);
   uint8_t (*read_holding_registers) (void);
   uint8_t (*read_input_registers) (void);
   uint8_t (*write_single_coil) (void);
   uint8_t (*write_single_register) (void);
   uint8_t (*write_multiple_coils) (void);
   uint8_t (*write_multiple_regsiters) (void);
   uint8_t (*write_mask_register) (void);
   uint8_t (*write_read_multiple_regsiters) (void);
} MB_FUNCTIONS_T;

typedef struct {
    uint16_t  transaction_id;
    uint16_t  protocol_id;
    uint16_t  length;
    uint8_t   unit_id;
    uint8_t   function_code; // this does not really belong to the MBAP, but it is common to all requests
} MBAP_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  StartAddr;
  uint16_t  NumCoils;
} READ_COILS_REQUEST_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  StartAddr;
  uint16_t  NumInputs;
} READ_DISCRETE_INPUTS_REQUEST_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  StartAddr;
  uint16_t  NumRegisters;
} READ_HOLDING_REGISTERS_REQUEST_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  StartAddr;
  uint16_t  NumRegisters;
} READ_INPUT_REGISTERS_REQUEST_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  Addr;
  uint16_t  Value;
} WRITE_SINGLE_COIL_REQUEST_T, WRITE_SINGLE_COIL_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  Addr;
  uint16_t  Value;
} WRITE_SINGLE_REGISTER_REQUEST_T, WRITE_SINGLE_REGISTER_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;       //  8 bytes
  uint16_t  StartAddr;  //  2 bytes
  uint16_t  NumCoils;   //  2 bytes
  uint8_t   Count;      //  1 byte
  uint8_t   Value[242]; //  242 bytes (1936 coils). Modbus spec is 246 x 8Bit values = 1968 coils
                        //  Total 255 bytes
} WRITE_MULTIPLE_COILS_REQUEST_T, WRITE_MULTIPLE_COILS_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;       //  8 bytes
  uint16_t  StartAddr;  //  2 bytes
  uint16_t  NumRegs;    //  2 bytes
  uint8_t   Count;      //  1 byte
  uint16_t  Value[121]; //  242 bytes (= 2 x 121) ; Modbus spec is 123 x 16Bit registers
                        //  Total 255 bytes + (2 * 121 =)242 bytes = 255 bytes
} WRITE_MULTIPLE_REGISTERS_REQUEST_T, WRITE_MULTIPLE_REGISTERS_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  RdStartAddr;
  uint16_t  RdNumRegs;
  uint16_t  WrStartAddr;
  uint16_t  WrNumRegs;
  uint8_t   Count;
  uint16_t  Value[119];  // 238 bytes (= 2 x 119)  ; Modbus spec is READ 125 (0x7D) x 16Bit = 250 bytes
                         //                        ; Modbus spec is WRITE 121 (0x79) x 16Bit = 242 bytes
} WRITE_READ_MULTIPLE_REGISTERS_REQUEST_T, WRITE_READ_MULTIPLE_REGISTERS_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;
  uint16_t  Addr;
  uint16_t  AndMask;
  uint16_t  OrMask;
} MASK_WRITE_REGISTER_REQUEST_T, MASK_WRITE_REGISTER_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;
  uint8_t   MEI_Type;
  uint8_t   ReadDevIdCode;
  uint8_t   ObjectId;
} READ_DEVICE_ID_REQUEST_T;

typedef struct {
	union {
        MBAP_T                                  mbap;
        READ_COILS_REQUEST_T                    RdCoils;
        READ_DISCRETE_INPUTS_REQUEST_T          RdDiscreteInputs;
        READ_HOLDING_REGISTERS_REQUEST_T        RdHoldingRegs;
        READ_INPUT_REGISTERS_REQUEST_T          RdInputRegs;
        WRITE_SINGLE_COIL_REQUEST_T             WrSingleCoil;
        WRITE_SINGLE_REGISTER_REQUEST_T         WrSingleReg;
        WRITE_MULTIPLE_COILS_REQUEST_T          WrMultCoils;
        WRITE_MULTIPLE_REGISTERS_REQUEST_T      WrMultRegs;
        WRITE_READ_MULTIPLE_REGISTERS_REQUEST_T WrRdMultRegs;
        MASK_WRITE_REGISTER_REQUEST_T           MaskWrReg;
        READ_DEVICE_ID_REQUEST_T                RdDeviceId;
        uint8_t packet[255];
	}u;
}MB_REQUEST_T;

typedef struct {
  MBAP_T    mbap;
  uint8_t   code;
} ERROR_CODE_RESPONSE_T;

typedef struct {
  MBAP_T   mbap;        //  8 bytes
  uint8_t  Count;       //  1 byte
  uint8_t  Status[246]; //  246 bytes => 1968 coils (= 8 x 246) ; Modbus spec is 2000 (0x07D0) / 8 = 250 x 8Bit
                        //  Total 255 bytes
} READ_COILS_RESPONSE_T, READ_DISCRETE_INPUTS_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;       //  8 bytes
  uint8_t   Count;      //  1 byte
  uint16_t  Value[123]; //  246 bytes (= 2 x 123) ; Modbus spec is 125 (0x7D) x 2 = 250 x 8Bit
                        //  Total 255 bytes
} READ_HOLDING_REGISTERS_RESPONSE_T, READ_INPUT_REGISTERS_RESPONSE_T;

typedef struct {
  MBAP_T    mbap;
  uint8_t   MEI_Type;
  uint8_t   ReadDevIdCode;
  uint8_t   ConformityLevel;
  uint8_t   MoreFollows;
  uint8_t   NextObjectId;
  uint8_t   NumOfObjects;
  uint8_t   Objects[241];
} READ_DEVICE_ID_RESPONSE_T;

typedef struct {
    uint8_t length;
    union {
        MBAP_T                                      mbap;
        ERROR_CODE_RESPONSE_T                       error;
        READ_COILS_RESPONSE_T                       RdCoils;
        READ_DISCRETE_INPUTS_RESPONSE_T             RdDiscreteInputs;
        READ_INPUT_REGISTERS_RESPONSE_T             RdInputRegs;
        READ_HOLDING_REGISTERS_RESPONSE_T           RdHoldingRegs;
        WRITE_SINGLE_COIL_RESPONSE_T                WrSingleCoil;
        WRITE_SINGLE_REGISTER_RESPONSE_T            WrSingleReg;
        WRITE_MULTIPLE_COILS_RESPONSE_T             WrMultCoils;
        WRITE_MULTIPLE_REGISTERS_RESPONSE_T         WrMultRegs;
        WRITE_READ_MULTIPLE_REGISTERS_RESPONSE_T    WrRdMultRegs;
        MASK_WRITE_REGISTER_RESPONSE_T              MaskWrReg;
        READ_DEVICE_ID_RESPONSE_T                   RdDeviceId;
        uint8_t packet[255];
    }u;
} MB_RESPONSE_T;

typedef struct {
    uint8_t         objects_size;
    uint8_t*        objects;
    struct {
        uint32_t    ip;
        uint16_t    port;
        uint8_t     ch;
        uint8_t     rxlen;
    } src_socket;
    struct {
        uint32_t    ip;
        uint16_t    port;
        uint8_t     ch;
        uint8_t     noheader;
    } dst_socket;
    MB_FUNCTIONS_T  f;
    MB_REQUEST_T    request;
    MB_RESPONSE_T   response;
} MODBUS_T;

extern MODBUS_T modbus;

static inline uint16_t u16swap(uint16_t val)
{
    return ((val >> 8) & 0x00FF) | ((val << 8) & 0xFF00);
}

extern void mb_register_function(uint8_t func_code, uint8_t(*func)(void));
extern void mb_init(uint8_t * mb_objects, uint8_t mb_objects_size, uint8_t socket);
extern void mb_set_data_length(uint16_t len);
extern void mb_handler(void);

#endif // MB_H_INCLUDED
