#include "vmdevice.h"
#include <stdint.h>
#include <math.h>
#include "i2c.h"
#include "ds1621.h"

#define CMD_STOP_CONVERT    0x22
#define CMD_ACCESS_TH       0xA1
#define CMD_ACCESS_TL       0xA2
#define CMD_ACCESS_CONFIG   0xAC
#define CMD_READ_COUNTER    0xA8
#define CMD_READ_SLOPE      0xA9
#define CMD_READ_TEMP       0xAA
#define CMD_START_CONVERT   0xEE

float uitemp_to_float(uint16_t uitemp)
{
    int ctemp;
    float ftemp;

    ctemp = (int) ((uitemp >> 8) & 0x00FF);
    ftemp = (float) ctemp;

    if ((uitemp & 0x0080) == 0x0080) {
        ftemp += 0.5;
    }
    return(ftemp);
}

uint16_t ftemp_to_u16(float ftemp)
{
    float fval;
    float ffrac;
    uint16_t lsb = 0x0000;
    uint16_t result;

    fval = ((int16_t)ftemp);
    ffrac = fabsf((ftemp - fval) * 10);

    if (ffrac >= 5) {
        if (ftemp < 0) {
            ftemp -= 0.5;
            fval = ((int16_t)ftemp);
        }
        lsb = 0x0080;
    }
    result = ((((int16_t) fval) << 8) & 0xFF00) | lsb;

    return(result);
}

uint8_t ds1621_start_convert(uint8_t slvaddr)
{
    return(i2c_write_8bit(slvaddr, CMD_START_CONVERT) == I2C_RESULT_OK);
}

uint8_t ds1621_stop_convert(uint8_t slvaddr)
{
    return(i2c_write_8bit(slvaddr, CMD_STOP_CONVERT) == I2C_RESULT_OK);
}

uint8_t ds1621_get_config(uint8_t slvaddr, uint8_t * config)
{
    return(i2c_reg8_rd8(slvaddr, CMD_ACCESS_CONFIG, config) == I2C_RESULT_OK);
}

uint8_t ds1621_set_config(uint8_t slvaddr, uint8_t config)
{
    return(i2c_reg8_wr8(slvaddr, CMD_ACCESS_CONFIG, config) == I2C_RESULT_OK);
}

uint8_t ds1621_read_temp(uint8_t slvaddr, uint16_t * temp16)
{
    return(i2c_reg8_rd16(slvaddr, CMD_READ_TEMP, temp16) == I2C_RESULT_OK);
}

uint8_t ds1621_set_TH(uint8_t slvaddr, uint16_t th)
{
    return(i2c_reg8_wr16(slvaddr, CMD_ACCESS_TH, th) == I2C_RESULT_OK);
}

uint8_t ds1621_set_TL(uint8_t slvaddr, uint16_t tl)
{
    return(i2c_reg8_wr16(slvaddr, CMD_ACCESS_TL, tl) == I2C_RESULT_OK);
}

uint8_t ds1621_get_TH(uint8_t slvaddr, uint16_t * th)
{
    return (i2c_reg8_rd16(slvaddr, CMD_ACCESS_TH, th) == I2C_RESULT_OK);
}

uint8_t ds1621_get_TL(uint8_t slvaddr, uint16_t * tl)
{
    return (i2c_reg8_rd16(slvaddr, CMD_ACCESS_TL, tl) == I2C_RESULT_OK);
}

uint8_t ds1621_read_counter(uint8_t slvaddr, uint8_t * counter)
{
    return(i2c_reg8_rd8(slvaddr, CMD_READ_COUNTER, counter) == I2C_RESULT_OK);
}

uint8_t ds1621_read_slope(uint8_t slvaddr, uint8_t * slope)
{
    return(i2c_reg8_rd8(slvaddr, CMD_READ_SLOPE, slope) == I2C_RESULT_OK);
}

uint8_t ds1621_read_hrtemp(uint8_t slvaddr, float * ftemp)
{
    uint8_t counter;
    uint8_t slope;
    uint16_t uitemp;
    int8_t itemp;
    uint8_t ret = 0;

    *ftemp = 0;

    if (ds1621_read_temp(slvaddr, &uitemp)) {
        if (ds1621_read_counter(slvaddr, &counter)) {
            if (ds1621_read_slope(slvaddr, &slope)) {
                itemp = ((uitemp >> 8) & 0x00FF);
                *ftemp = (float) itemp;

                if (slope == 0) {
                    if ((uitemp & 0x0080) == 0x0080) {
                        *ftemp += 0.5;
                    }
                } else {
                    *ftemp = ((*ftemp) - 0.25) + (((float) slope - (float) counter) / (float) slope);
                }
                ret = 1;
            }
        }
    }
    return(ret);
}

uint8_t init_ds1621(uint8_t slvaddr, float tl, float th)
{
    uint8_t ret;
    uint16_t tl_tmp = 0;
    uint16_t th_tmp = 0;
    uint16_t l_tl;
    uint16_t l_th;

    ds1621_stop_convert(slvaddr);

    l_tl = ftemp_to_u16(tl);
    l_th = ftemp_to_u16(th);

    ret = ds1621_get_TH(slvaddr, &th_tmp);
    ret &= ds1621_get_TH(slvaddr, &tl_tmp);

    if (ret) {
        if (th_tmp != l_th) {
                ret &= ds1621_set_TH(slvaddr, l_th);
        }
        if (tl_tmp != tl) {
                ret &= ds1621_set_TL(slvaddr, l_tl);
        }
    }

    //Clear THF and THF
    //Continuous temperature conversion
    //Output Polarity Bit. 0 = active low
    ret &= ds1621_set_config(slvaddr, 0x00);
    ret &= ds1621_start_convert(slvaddr);

    return(ret);
}

uint8_t ds1621_reset_flags(uint8_t slvaddr)
{
    uint8_t config;
    uint8_t ret;

    ret = ds1621_get_config(slvaddr, &config);
    if (ret) {
        ret = ds1621_set_config(slvaddr, config & (~(DS1621_THF | DS1621_TLF)));
    }
    return(ret);
}



