#include "vmdevice.h"
#include <stdint.h>
#include <stdio.h>
#include "net.h"
#include "mb.h"

MODBUS_T modbus;

/*
*   Default function returns ModBus error code (0x01) ILLEGAL FUNCTION
*   if a function is not implemented and not registered.
*
*/
uint8_t default_mb_function(void) {
    return(MB_ERR_ILLEGAL_FUNCTION);
}

/*
*   Registers a ModBus function
*   func_code: 8bit ModBus function code
*   *func: pointer to the functon
*/
void mb_register_function(uint8_t func_code, uint8_t(*func)(void))
{
    uint8_t (*pfunc)(void);

    if(!func) {
        pfunc = default_mb_function;
    } else {
        pfunc = func;
    }

    switch(func_code) {
    case MB_CMD_READ_COILS:                     modbus.f.read_coils = pfunc; break;
    case MB_CMD_READ_DISCRETE_INPUTS:           modbus.f.read_discrete_inputs = pfunc; break;
    case MB_CMD_READ_HOLDING_REGISTERS:         modbus.f.read_holding_registers = pfunc; break;
    case MB_CMD_READ_INPUT_REGISTERS:           modbus.f.read_input_registers = pfunc; break;
    case MB_CMD_WRITE_SINGLE_COIL:              modbus.f.write_single_coil = pfunc; break;
    case MB_CMD_WRITE_SINGLE_REGISTER:          modbus.f.write_single_register = pfunc; break;
    case MB_CMD_WRITE_MULTIPLE_COILS:           modbus.f.write_multiple_coils = pfunc; break;
    case MB_CMD_WRITE_MULTIPLE_REGISTERS:       modbus.f.write_multiple_regsiters = pfunc; break;
    case MB_CMD_WRITE_MASK_REGISTER:            modbus.f.write_mask_register = pfunc; break;
    case MB_CMD_WRITE_READ_MULTIPLE_REGISTERS:  modbus.f.write_read_multiple_regsiters = pfunc; break;
    }
}

/*
*   Prior ModBus can be used, mb:init must be call to initialize ModBus
*   *mb_objects: pointer to the object containing the device identification (VendorName, ProductCode and Revision)
*   mb_objects_size: The size of the object [Bytes]
*   socket: the socket number ModBus will use: 3 for TCP/IP and 4 for UDP
*/
void mb_init(uint8_t * mb_objects, uint8_t mb_objects_size, uint8_t socket)
{
    // Socket 3 -> (VM) TCP/IP
    // Socket 4 -> (VM) UDP
    if ((socket != 3) && (socket != 4)) socket = 3;

    modbus.dst_socket.ch        = socket;
    modbus.dst_socket.ip        = 0;
    modbus.dst_socket.port      = 0;
    modbus.dst_socket.noheader  = 1;

    modbus.objects_size = mb_objects_size;
    modbus.objects = mb_objects;

    mb_register_function(MB_CMD_READ_COILS, NULL);
    mb_register_function(MB_CMD_READ_DISCRETE_INPUTS, NULL);
    mb_register_function(MB_CMD_READ_HOLDING_REGISTERS, NULL);
    mb_register_function(MB_CMD_READ_INPUT_REGISTERS, NULL);
    mb_register_function(MB_CMD_WRITE_SINGLE_COIL, NULL);
    mb_register_function(MB_CMD_WRITE_SINGLE_REGISTER, NULL);
    mb_register_function(MB_CMD_WRITE_MULTIPLE_COILS, NULL);
    mb_register_function(MB_CMD_WRITE_MULTIPLE_REGISTERS, NULL);
    mb_register_function(MB_CMD_WRITE_MASK_REGISTER, NULL);
    mb_register_function(MB_CMD_WRITE_READ_MULTIPLE_REGISTERS, NULL);
}
/*
*   This function stores the ModBus request to the Variable (structure) "modbus"
*   before processing the request.
*/
void mb_backup_request(void)
{
    uint8_t i;

    modbus.src_socket.ip    = RXIP;
    modbus.src_socket.port  = RXPORT;
    modbus.src_socket.ch    = RXCH;
    modbus.src_socket.rxlen = NETRXLEN;

    for(i=0; i<modbus.src_socket.rxlen; i++) {
        modbus.request.u.packet[i] = NETRXBUF[i];
    }
}

/*
*   This function copies request data to response data
*/
void mb_copy_mbap(void)
{
    modbus.response.u.mbap.transaction_id   = modbus.request.u.mbap.transaction_id;
    modbus.response.u.mbap.protocol_id      = modbus.request.u.mbap.protocol_id;
    modbus.response.u.mbap.length           = 0; // to be set later
    modbus.response.u.mbap.unit_id          = modbus.request.u.mbap.unit_id;
    modbus.response.u.mbap.function_code    = modbus.request.u.mbap.function_code;
}

void mb_set_data_length(uint16_t len)
{
    len += 2;  // plus function_code and unit_id bytes
    modbus.response.u.mbap.length = u16swap(len);
    modbus.response.length = len + 6; // the length of the whole packet including MBAP needed to set NETTXLEN
}

void mb_set_exception_code(uint8_t err_code)
{
    modbus.response.u.mbap.function_code += 0x80;
    modbus.response.u.error.code = err_code;
    mb_set_data_length(1);  // only one byte -> the err_code
}

uint8_t mb_read_device_information(void)
{
    uint8_t ret;
    uint8_t i;

    if (modbus.request.u.RdDeviceId.MEI_Type != 0x0E)
        return(MB_ERR_ILLEGAL_FUNCTION);

    if (modbus.request.u.RdDeviceId.ObjectId != 0x00)
        return(MB_ERR_ILLEGAL_DATA_ADDRESS);

    if (modbus.request.u.RdDeviceId.ReadDevIdCode != 0x01)
        return(MB_ERR_ILLEGAL_DATA_VALUE);

    //6 bytes
    modbus.response.u.RdDeviceId.MEI_Type           = 0x0E;
    modbus.response.u.RdDeviceId.ReadDevIdCode      = 0x01;
    modbus.response.u.RdDeviceId.ConformityLevel    = 0x01;
    modbus.response.u.RdDeviceId.MoreFollows        = 0x00;
    modbus.response.u.RdDeviceId.NextObjectId       = 0x00;
    modbus.response.u.RdDeviceId.NumOfObjects       = 0x03;

    for(i=0;i<modbus.objects_size;i++) {
        modbus.response.u.RdDeviceId.Objects[i] = modbus.objects[i];
    }

    //total length = 6 bytes + size of objects
    mb_set_data_length(6 + modbus.objects_size);
    ret = MB_OK;
    return(ret);
}

void mb_handler(void)
{
    uint8_t ret;

    //copy data received from a Modbus master
    mb_backup_request();

    //clear the network flag to allow reception of new data
    FNET = 0;

    //copy MBAP data from REQUEST to RESPONSE
    mb_copy_mbap();

    switch(modbus.response.u.mbap.function_code) {
    case MB_CMD_READ_COILS:
        ret = modbus.f.read_coils();
        break;
    case MB_CMD_READ_DISCRETE_INPUTS:
        ret = modbus.f.read_discrete_inputs();
        break;
    case MB_CMD_READ_HOLDING_REGISTERS:
        ret = modbus.f.read_holding_registers();
        break;
     case MB_CMD_READ_INPUT_REGISTERS:
        ret = modbus.f.read_input_registers();
        break;
    case MB_CMD_WRITE_SINGLE_COIL:
        ret = modbus.f.write_single_coil();
        break;
    case MB_CMD_WRITE_SINGLE_REGISTER:
        ret = modbus.f.write_single_register();
        break;
    case MB_CMD_WRITE_MULTIPLE_COILS:
        ret = modbus.f.write_multiple_coils();
        break;
    case MB_CMD_WRITE_MULTIPLE_REGISTERS:
        ret = modbus.f.write_multiple_regsiters();
        break;
    case MB_CMD_WRITE_MASK_REGISTER:
        ret = modbus.f.write_mask_register();
        break;
    case MB_CMD_WRITE_READ_MULTIPLE_REGISTERS:
        ret = modbus.f.write_read_multiple_regsiters();
        break;
    case MB_CMD_READ_DEVICE_INFORMATION:
        ret = mb_read_device_information();
        break;
    default:
        ret = MB_ERR_ILLEGAL_FUNCTION;
    }

    if (ret != MB_OK) {
        mb_set_exception_code(ret);
    }

    if (ret != MB_ERR_UNKNOWN) {
        net_sendbytes(modbus.response.u.packet, modbus.response.length, modbus.dst_socket.ch);
    }
}
